@@ -3,38 +3,48 @@
 #define _XOPEN_SOURCE 600  /* Make sure strdup() is in <string.h> */
 #define _BSD_SOURCE  /* Make sure setgroups()is in <grp.h> */
 
+#include "xmlrpc_config.h"
+
 #include <assert.h>
 #include <stdio.h>
 #include <stdlib.h>
+#if MSVCRT
+  #include <process.h>
+#else
+  #include <unistd.h>
+#endif
 #include <string.h>
 #include <time.h>
 #include <errno.h>
-#ifndef WIN32
+#if !MSVCRT
   #include <grp.h>
 #endif
 
-#include "xmlrpc_config.h"
 #include "bool.h"
 #include "girmath.h"
 #include "mallocvar.h"
 #include "xmlrpc-c/string_int.h"
 #include "xmlrpc-c/sleep_int.h"
+#include "xmlrpc-c/lock.h"
+#include "xmlrpc-c/lock_platform.h"
 
 #include "xmlrpc-c/abyss.h"
 #include "trace.h"
+#include "thread.h"
 #include "session.h"
 #include "file.h"
 #include "conn.h"
 #include "chanswitch.h"
 #include "channel.h"
 #include "socket.h"
-#ifdef WIN32
+#if MSVCRT
   #include "socket_win.h"
 #else
   #include "socket_unix.h"
 #endif
 #include "http.h"
 #include "handler.h"
+#include "sessionReadRequest.h"
 
 #include "server.h"
 
@@ -75,7 +85,7 @@ ServerResetTerminate(TServer * const serverP) {
 
 static void
 initUnixStuff(struct _TServer * const srvP) {
-#ifndef WIN32
+#if !MSVCRT
     srvP->pidfileP = NULL;
     srvP->uid = srvP->gid = -1;
 #endif
@@ -83,27 +93,24 @@ initUnixStuff(struct _TServer * const srvP) {
 
 
 
-static bool
-logOpen(struct _TServer * const srvP) {
+static void
+logOpen(struct _TServer * const srvP,
+        const char **     const errorP) {
 
     bool success;
 
     success = FileOpenCreate(&srvP->logfileP, srvP->logfilename,
                              O_WRONLY | O_APPEND);
     if (success) {
-        bool success;
-        success = MutexCreate(&srvP->logmutexP);
-        if (success)
-            srvP->logfileisopen = TRUE;
-        else
-            TraceMsg("Can't create mutex for log file");
+        srvP->logLockP = xmlrpc_lock_create();
+        *errorP = NULL;
 
-        if (!success)
+        srvP->logfileisopen = true;
+            
+        if (*errorP)
             FileClose(srvP->logfileP);
     } else
-        TraceMsg("Can't open log file '%s'", srvP->logfilename);
-
-    return success;
+        xmlrpc_asprintf(errorP, "Can't open log file '%s'", srvP->logfilename);
 }
 
 
@@ -113,8 +120,48 @@ logClose(struct _TServer * const srvP) {
 
     if (srvP->logfileisopen) {
         FileClose(srvP->logfileP);
-        MutexDestroy(srvP->logmutexP);
-        srvP->logfileisopen = FALSE;
+        srvP->logLockP->destroy(srvP->logLockP);
+        srvP->logfileisopen = false;
+    }
+}
+
+
+
+static void
+setupTrace(struct _TServer * const srvP) {
+
+    srvP->tracer.traceIsActive = (getenv("ABYSS_TRACE_SERVER") != NULL);
+
+    if (srvP->tracer.traceIsActive)
+        fprintf(stderr, "Abyss server will trace "
+                "basic server activity "
+                "because of ABYSS_TRACE_SERVER environment variable\n");
+}
+
+
+
+static void
+tracev(const char * const fmt,
+       va_list            argptr) {
+
+    vfprintf(stderr, fmt, argptr);
+
+    fprintf(stderr, "\n");
+}
+
+
+
+static void
+trace(struct Tracer * const tracerP,
+      const char *    const fmt,
+      ...) {
+
+    if (tracerP->traceIsActive) {
+        va_list argptr;
+
+        va_start(argptr, fmt);
+        tracev(fmt, argptr);
+        va_end(argptr);
     }
 }
 
@@ -131,21 +178,22 @@ initChanSwitchStuff(struct _TServer * const srvP,
 
     if (chanSwitchP) {
         *errorP = NULL;
-        srvP->serverAcceptsConnections = TRUE;
+        srvP->serverAcceptsConnections = true;
         srvP->chanSwitchP = chanSwitchP;
         srvP->weCreatedChanSwitch = !userChanSwitch;
     } else if (noAccept) {
         *errorP = NULL;
-        srvP->serverAcceptsConnections = FALSE;
+        srvP->serverAcceptsConnections = false;
         srvP->chanSwitchP = NULL;
-        srvP->weCreatedChanSwitch = FALSE;
+        srvP->weCreatedChanSwitch = false;
     } else {
         *errorP = NULL;
-        srvP->serverAcceptsConnections = TRUE;
+        srvP->serverAcceptsConnections = true;
         srvP->chanSwitchP = NULL;
-        srvP->weCreatedChanSwitch = FALSE;
+        srvP->weCreatedChanSwitch = false;
         srvP->port = port;
     }
+    srvP->readyToAccept = false;
 }
 
 
@@ -157,7 +205,9 @@ createServer(struct _TServer ** const srvPP,
              bool               const userChanSwitch,
              unsigned short     const portNumber,             
              const char **      const errorP) {
-
+/*----------------------------------------------------------------------------
+   Create a server object and return it as *srvPP.
+-----------------------------------------------------------------------------*/
     struct _TServer * srvP;
 
     MALLOCVAR(srvP);
@@ -166,6 +216,8 @@ createServer(struct _TServer ** const srvPP,
         xmlrpc_asprintf(errorP,
                         "Unable to allocate space for server descriptor");
     } else {
+        setupTrace(srvP);
+
         srvP->terminationRequested = false;
 
         initChanSwitchStuff(srvP, noAccept, chanSwitchP, userChanSwitch,
@@ -185,15 +237,18 @@ createServer(struct _TServer ** const srvPP,
                 srvP->keepalivetimeout = 15;
                 srvP->keepalivemaxconn = 30;
                 srvP->timeout          = 15;
-                srvP->advertise        = TRUE;
-                srvP->useSigchld       = FALSE;
+                srvP->advertise        = true;
+                srvP->useSigchld       = false;
                 srvP->uriHandlerStackSize = 0;
-            
+                srvP->maxConn          = 15;
+                srvP->maxConnBacklog   = 15;
+                srvP->maxSessionMem    = 0;
+
                 initUnixStuff(srvP);
 
                 ListInitAutoFree(&srvP->handlers);
 
-                srvP->logfileisopen = FALSE;
+                srvP->logfileisopen = false;
                 
                 *errorP = NULL;
 
@@ -237,9 +292,15 @@ ServerCreate(TServer *       const serverP,
              xmlrpc_uint16_t const portNumber,
              const char *    const filesPath,
              const char *    const logFileName) {
+/*----------------------------------------------------------------------------
+   Create a server that will listen for and accept client connections on TCP
+   port 'portNumber'.  Those connections will be raw TCP connections.
 
-    bool const noAcceptFalse = FALSE;
-    bool const userChanSwitchFalse = FALSE;
+   Set *serverP to represent that server.
+-----------------------------------------------------------------------------*/
+
+    bool const noAcceptFalse = false;
+    bool const userChanSwitchFalse = false;
 
     bool success;
     const char * error;
@@ -251,9 +312,9 @@ ServerCreate(TServer *       const serverP,
     if (error) {
         TraceMsg(error);
         xmlrpc_strfree(error);
-        success = FALSE;
+        success = false;
     } else {
-        success = TRUE;
+        success = true;
     
         setNamePathLog(serverP, name, filesPath, logFileName);
     }
@@ -268,7 +329,7 @@ createSwitchFromOsSocket(TOsSocket      const osSocket,
                          TChanSwitch ** const chanSwitchPP,
                          const char **  const errorP) {
 
-#ifdef WIN32
+#if MSVCRT
     ChanSwitchWinCreateWinsock(osSocket, chanSwitchPP, errorP);
 #else
     ChanSwitchUnixCreateFd(osSocket, chanSwitchPP, errorP);
@@ -283,7 +344,7 @@ createChannelFromOsSocket(TOsSocket     const osSocket,
                           void **       const channelInfoPP,
                           const char ** const errorP) {
 
-#ifdef WIN32
+#if MSVCRT
     ChannelWinCreateWinsock(osSocket, channelPP,
                             (struct abyss_win_chaninfo**)channelInfoPP,
                             errorP);
@@ -302,7 +363,12 @@ ServerCreateSocket(TServer *    const serverP,
                    TOsSocket    const socketFd,
                    const char * const filesPath,
                    const char * const logFileName) {
+/*----------------------------------------------------------------------------
+   Create a server that will listen for and accept client connections on
+   socket 'socketFd'.  Those connections will be raw TCP connections.
 
+   Set *serverP to represent that server.
+-----------------------------------------------------------------------------*/
     bool success;
     TChanSwitch * chanSwitchP;
     const char * error;
@@ -311,11 +377,11 @@ ServerCreateSocket(TServer *    const serverP,
 
     if (error) {
         TraceMsg(error);
-        success = FALSE;
+        success = false;
         xmlrpc_strfree(error);
     } else {
-        bool const noAcceptFalse = FALSE;
-        bool const userChanSwitchFalse = FALSE;
+        bool const noAcceptFalse = false;
+        bool const userChanSwitchFalse = false;
 
         const char * error;
 
@@ -325,10 +391,10 @@ ServerCreateSocket(TServer *    const serverP,
 
         if (error) {
             TraceMsg(error);
-            success = FALSE;
+            success = false;
             xmlrpc_strfree(error);
         } else {
-            success = TRUE;
+            success = true;
             
             setNamePathLog(serverP, name, filesPath, logFileName);
         }
@@ -346,9 +412,15 @@ ServerCreateNoAccept(TServer *    const serverP,
                      const char * const name,
                      const char * const filesPath,
                      const char * const logFileName) {
+/*----------------------------------------------------------------------------
+   Create a server that will not get client connections; the user will have to
+   get the client connections and use server methods that take an existing
+   connection as an argument.
 
-    bool const noAcceptTrue = TRUE;
-    bool const userChanSwitchFalse = FALSE;
+   Set *serverP to represent that server.
+-----------------------------------------------------------------------------*/
+    bool const noAcceptTrue = true;
+    bool const userChanSwitchFalse = false;
 
     bool success;
     const char * error;
@@ -359,10 +431,10 @@ ServerCreateNoAccept(TServer *    const serverP,
 
     if (error) {
         TraceMsg(error);
-        success = FALSE;
+        success = false;
         xmlrpc_strfree(error);
     } else {
-        success = TRUE;
+        success = true;
         
         setNamePathLog(serverP, name, filesPath, logFileName);
     }
@@ -375,9 +447,16 @@ void
 ServerCreateSwitch(TServer *     const serverP,
                    TChanSwitch * const chanSwitchP,
                    const char ** const errorP) {
-    
-    bool const noAcceptFalse = FALSE;
-    bool const userChanSwitchTrue = TRUE;
+/*----------------------------------------------------------------------------
+   Create a server that will get client connections from channel switch
+   *chanSwitchP.  Those connections would typically be raw TCP connections,
+   but could be anything that can be abstracted by a Channel object.  For
+   example, an SSL encrypted TCP connection.
+
+   Set *serverP to represent that server.
+-----------------------------------------------------------------------------*/
+    bool const noAcceptFalse = false;
+    bool const userChanSwitchTrue = true;
 
     assert(serverP);
     assert(chanSwitchP);
@@ -532,6 +611,38 @@ ServerSetMimeType(TServer *  const serverP,
 
 
 
+void
+ServerSetMaxConn(TServer *    const serverP,
+                 unsigned int const maxConn) {
+
+    if (maxConn > 0) {
+        serverP->srvP->maxConn = maxConn;
+    }
+}
+
+
+
+void
+ServerSetMaxConnBacklog(TServer *    const serverP,
+                        unsigned int const maxConnBacklog) {
+
+    if (maxConnBacklog > 0) {
+        serverP->srvP->maxConnBacklog = maxConnBacklog;
+    }
+}
+
+
+
+void
+ServerSetMaxSessionMem(TServer * const serverP,
+                       size_t    const size) {
+
+    if (size > 0)
+        serverP->srvP->maxSessionMem = size;
+}
+
+
+
 static URIHandler2
 makeUriHandler2(const struct uriHandler * const handlerP) {
 
@@ -555,7 +666,7 @@ runUserHandler(TSession *        const sessionP,
     abyss_bool handled;
     int i;
     
-    for (i = srvP->handlers.size-1, handled = FALSE;
+    for (i = srvP->handlers.size-1, handled = false;
          i >= 0 && !handled;
          --i) {
         const struct uriHandler * const handlerP = srvP->handlers.item[i];
@@ -606,17 +717,31 @@ handleReqInvalidURI(TSession * const sessionP) {
 
 
 static void
-processRequestFromClient(TConn *  const connectionP,
-                         bool     const lastReqOnConn,
-                         uint32_t const timeout,
-                         bool *   const keepAliveP) {
+traceRequestStart(struct Tracer * const tracerP,
+                  TSession *      const sessionP) {
+
+    if (sessionP->requestInfo.uri) {
+        trace(tracerP, "Processing request with URI '%s', method %s",
+              sessionP->requestInfo.uri,
+              HTTPMethodName(sessionP->requestInfo.method));
+    }
+}
+
+
+
+static void
+processRequestFromClient(TConn *         const connectionP,
+                         bool            const lastReqOnConn,
+                         uint32_t        const timeout,
+                         struct Tracer * const tracerP,
+                         bool *          const keepAliveP) {
 /*----------------------------------------------------------------------------
    Get and execute one HTTP request from client connection *connectionP,
    through the connection buffer.  I.e. Some of the request may already be in
    the connection buffer, and we may leave some of later requests in the
    connection buffer.
 
-   In fact, due to timing considerations, we assume the client has begun
+   In fact, because of timing considerations, we assume the client has begun
    sending the request, which as a practical matter means Caller has already
    deposited some of it in the connection buffer.
 
@@ -638,20 +763,21 @@ processRequestFromClient(TConn *  const connectionP,
     const char * error;
     uint16_t httpErrorCode;
 
-    RequestInit(&session, connectionP);
+    SessionInit(&session, connectionP);
 
     session.serverDeniesKeepalive = lastReqOnConn;
         
-    RequestRead(&session, timeout, &error, &httpErrorCode);
+    SessionReadRequest(&session, timeout, &error, &httpErrorCode);
 
     if (error) {
         ResponseStatus(&session, httpErrorCode);
         ResponseError2(&session, error);
         xmlrpc_strfree(error);
     } else {
+        traceRequestStart(tracerP, &session);
         if (session.version.major >= 2)
             handleReqTooNewHttpVersion(&session);
-        else if (!RequestValidURI(&session))
+        else if (!HTTPRequestHasValidUri(&session))
             handleReqInvalidURI(&session);
         else
             runUserHandler(&session, connectionP->server->srvP);
@@ -660,7 +786,7 @@ processRequestFromClient(TConn *  const connectionP,
     assert(session.status != 0);
 
     if (session.responseStarted)
-        HTTPWriteEndChunk(&session);
+        ResponseWriteEnd(&session);
     else
         ResponseError(&session);
 
@@ -668,7 +794,7 @@ processRequestFromClient(TConn *  const connectionP,
 
     SessionLog(&session);
 
-    RequestFree(&session);
+    SessionTerm(&session);
 }
 
 
@@ -689,8 +815,12 @@ serverFunc(void * const userHandle) {
     bool connectionDone;
         /* No more need for this HTTP connection */
 
+    trace(&srvP->tracer,
+          "Thread starting to handle requests on a new connection.  "
+          "PID = %d", XMLRPC_GETPID());
+
     requestCount = 0;
-    connectionDone = FALSE;
+    connectionDone = false;
 
     while (!connectionDone) {
         bool timedOut, eof;
@@ -704,34 +834,43 @@ serverFunc(void * const userHandle) {
         ConnRead(connectionP, srvP->keepalivetimeout,
                  &timedOut, &eof, &readError);
 
-        if (readError) {
+        if (srvP->terminationRequested) {
+            connectionDone = true;
+        } else if (readError) {
             TraceMsg("Failed to read from Abyss connection.  %s", readError);
             xmlrpc_strfree(readError);
-            connectionDone = TRUE;
+            connectionDone = true;
         } else if (timedOut) {
-            connectionDone = TRUE;
+            connectionDone = true;
         } else if (eof) {
-            connectionDone = TRUE;
-        } else if (srvP->terminationRequested) {
-            connectionDone = TRUE;
+            connectionDone = true;
         } else {
             bool const lastReqOnConn =
                 requestCount + 1 >= srvP->keepalivemaxconn;
 
             bool keepalive;
+
+            trace(&srvP->tracer,
+                  "HTTP request %u at least partially received.  "
+                  "Receiving the rest and processing", requestCount);
             
             processRequestFromClient(connectionP, lastReqOnConn, srvP->timeout,
-                                     &keepalive);
+                                     &srvP->tracer, &keepalive);
+
+            trace(&srvP->tracer,
+                  "Done processing the HTTP request.  Keepalive = %s",
+                  keepalive ? "YES" : "NO");
             
             ++requestCount;
 
             if (!keepalive)
-                connectionDone = TRUE;
+                connectionDone = true;
             
             /**************** Must adjust the read buffer *****************/
             ConnReadInit(connectionP);
         }
     }
+    trace(&srvP->tracer, "PID %d done with connection", XMLRPC_GETPID());
 }
 
 
@@ -749,7 +888,7 @@ createSwitchFromPortNum(unsigned short const portNumber,
                         TChanSwitch ** const chanSwitchPP,
                         const char **  const errorP) {
 
-#ifdef WIN32
+#if MSVCRT
     ChanSwitchWinCreate(portNumber, chanSwitchPP, errorP);
 #else
     ChanSwitchUnixCreate(portNumber, chanSwitchPP, errorP);
@@ -777,7 +916,7 @@ createChanSwitch(struct _TServer * const srvP,
     } else {
         *errorP = NULL;
         
-        srvP->weCreatedChanSwitch = TRUE;
+        srvP->weCreatedChanSwitch = true;
         srvP->chanSwitchP         = chanSwitchP;
     }
 }
@@ -785,7 +924,8 @@ createChanSwitch(struct _TServer * const srvP,
 
 
 void
-ServerInit(TServer * const serverP) {
+ServerInit2(TServer *     const serverP,
+            const char ** const errorP) {
 /*----------------------------------------------------------------------------
    Initialize a server to accept connections.
 
@@ -795,45 +935,57 @@ ServerInit(TServer * const serverP) {
    user supplies the channels (TCP connections)).
 -----------------------------------------------------------------------------*/
     struct _TServer * const srvP = serverP->srvP;
-    const char * retError;
     
     if (!srvP->serverAcceptsConnections)
-        xmlrpc_asprintf(&retError,
+        xmlrpc_asprintf(errorP,
                         "ServerInit() is not valid on a server that doesn't "
                         "accept connections "
                         "(i.e. created with ServerCreateNoAccept)");
     else {
-        retError = NULL;  /* initial value */
+        *errorP = NULL;  /* initial value */
 
         if (!srvP->chanSwitchP) {
             const char * error;
             createChanSwitch(srvP, &error);
 
             if (error) {
-                xmlrpc_asprintf(&retError, "Unable to create a channel switch "
+                xmlrpc_asprintf(errorP, "Unable to create a channel switch "
                                 "for the server.  %s", error);
                 xmlrpc_strfree(error);
             }
         }
 
-        if (!retError) {
+        if (!*errorP) {
             const char * error;
 
             assert(srvP->chanSwitchP);
 
-            ChanSwitchListen(srvP->chanSwitchP, MAX_CONN, &error);
+            ChanSwitchListen(srvP->chanSwitchP, srvP->maxConnBacklog, &error);
 
             if (error) {
-                xmlrpc_asprintf(&retError,
+                xmlrpc_asprintf(errorP,
                                 "Failed to listen on bound socket.  %s",
                                 error);
                 xmlrpc_strfree(error);
+            } else {
+                srvP->readyToAccept = true;
             }
         }
     }
-    if (retError) {
-        TraceExit("ServerInit() failed.  %s", retError);
-        xmlrpc_strfree(retError);
+}
+
+
+
+void
+ServerInit(TServer * const serverP) {
+
+    const char * error;
+
+    ServerInit2(serverP, &error);
+
+    if (error) {
+        TraceExit("ServerInit() failed.  %s", error);
+        xmlrpc_strfree(error);
     }
 }
 
@@ -960,18 +1112,12 @@ waitForNoConnections(outstandingConnList * const outstandingConnListP) {
 
 
 static void
-waitForConnectionCapacity(outstandingConnList * const outstandingConnListP) {
+waitForConnectionCapacity(outstandingConnList * const outstandingConnListP,
+                          unsigned int          const maxConn) {
 /*----------------------------------------------------------------------------
-   Wait until there are fewer than the maximum allowed connections in
-   progress.
+   Wait until there are fewer than 'maxConn' connections in progress.
 -----------------------------------------------------------------------------*/
-    /* We need to make this number configurable.  Note that MAX_CONN (16) is
-       also the backlog limit on the TCP socket, and they really aren't
-       related.  As it stands, we can have 16 connections in progress inside
-       Abyss plus 16 waiting in the the channel switch.
-    */
-
-    while (outstandingConnListP->count >= MAX_CONN) {
+    while (outstandingConnListP->count >= maxConn) {
         freeFinishedConns(outstandingConnListP);
         if (outstandingConnListP->firstP)
             waitForConnectionFreed(outstandingConnListP);
@@ -980,7 +1126,29 @@ waitForConnectionCapacity(outstandingConnList * const outstandingConnListP) {
 
 
 
-#ifndef WIN32
+static void
+interruptChannels(outstandingConnList * const outstandingConnListP) {
+/*----------------------------------------------------------------------------
+   Get every thread that is waiting to read a request or write a response
+   for a connection to stop waiting.
+-----------------------------------------------------------------------------*/
+    TConn * connP;
+
+    for (connP = outstandingConnListP->firstP;
+         connP; connP = connP->nextOutstandingP) {
+
+        if (connP->finished) {
+            /* The connection couldn't be waiting on the channel, and the
+               channel probably doesn't even exit anymore.
+            */
+        } else 
+            ChannelInterrupt(connP->channelP);
+    }
+}
+
+
+
+#if !MSVCRT
 void
 ServerHandleSigchld(pid_t const pid) {
 
@@ -995,7 +1163,7 @@ ServerUseSigchld(TServer * const serverP) {
 
     struct _TServer * const srvP = serverP->srvP;
     
-    srvP->useSigchld = TRUE;
+    srvP->useSigchld = true;
 }
 
 
@@ -1023,55 +1191,98 @@ destroyChannel(void * const userHandle) {
 
 
 
+static void
+processNewChannel(TServer *             const serverP,
+                  TChannel *            const channelP,
+                  void *                const channelInfoP,
+                  outstandingConnList * const outstandingConnListP,
+                  const char **         const errorP) {
+
+    struct _TServer * const srvP = serverP->srvP;
+                      
+    TConn * connectionP;
+    const char * error;
+
+    freeFinishedConns(outstandingConnListP);
+            
+    trace(&srvP->tracer, "Waiting for there to be fewer than the maximum "
+          "%u sessions in progress",
+          srvP->maxConn);
+
+    waitForConnectionCapacity(outstandingConnListP, srvP->maxConn);
+            
+    ConnCreate(&connectionP, serverP, channelP, channelInfoP,
+               &serverFunc,
+               SERVER_FUNC_STACK + srvP->uriHandlerStackSize,
+               &destroyChannel, ABYSS_BACKGROUND,
+               srvP->useSigchld,
+               &error);
+    if (!error) {
+        addToOutstandingConnList(outstandingConnListP, connectionP);
+        ConnProcess(connectionP);
+        /* When connection is done (which could be later, courtesy of a
+           background thread), destroyChannel() will destroy *channelP.
+        */
+        *errorP = NULL;
+    } else {
+        xmlrpc_asprintf(
+            errorP, "Failed to create an Abyss connection.  %s", error);
+        xmlrpc_strfree(error);
+    }
+}
+
+
+
 static void
 acceptAndProcessNextConnection(
     TServer *             const serverP,
-    outstandingConnList * const outstandingConnListP) {
+    outstandingConnList * const outstandingConnListP,
+    const char **         const errorP) {
 
     struct _TServer * const srvP = serverP->srvP;
 
-    TConn * connectionP;
     const char * error;
     TChannel * channelP;
     void * channelInfoP;
+
+    trace(&srvP->tracer, "Waiting for a new channel from channel switch");
+
+    assert(srvP->readyToAccept);
+    assert(srvP->chanSwitchP);
         
     ChanSwitchAccept(srvP->chanSwitchP, &channelP, &channelInfoP, &error);
     
     if (error) {
-        TraceMsg("Failed to accept the next connection from a client "
-                 "at the channel level.  %s", error);
+        xmlrpc_asprintf(errorP,
+                        "Failed to accept the next connection from a client "
+                        "at the channel level.  %s", error);
         xmlrpc_strfree(error);
     } else {
         if (channelP) {
             const char * error;
 
-            freeFinishedConns(outstandingConnListP);
-            
-            waitForConnectionCapacity(outstandingConnListP);
-            
-            ConnCreate(&connectionP, serverP, channelP, channelInfoP,
-                       &serverFunc,
-                       SERVER_FUNC_STACK + srvP->uriHandlerStackSize,
-                       &destroyChannel, ABYSS_BACKGROUND,
-                       srvP->useSigchld,
-                       &error);
-            if (!error) {
-                addToOutstandingConnList(outstandingConnListP,
-                                         connectionP);
-                ConnProcess(connectionP);
-                /* When connection is done (which could be later, courtesy
-                   of a background thread), destroyChannel() will
-                   destroy *channelP.
-                */
-            } else {
-                TraceMsg("Failed to create an Abyss connection "
-                         "out of new channel %lx.  %s", channelP, error);
-                xmlrpc_strfree(error);
+            trace(&srvP->tracer, "Got a new channel from channel switch");
+
+            processNewChannel(serverP, channelP, channelInfoP,
+                              outstandingConnListP, &error);
+
+            if (error) {
+                xmlrpc_asprintf(errorP, "Failed to use new channel %lx",
+                                (unsigned long) channelP);
                 ChannelDestroy(channelP);
                 free(channelInfoP);
+            } else {
+                trace(&srvP->tracer,
+                      "successfully processed newly accepted channel");
+                /* Connection created above will destroy *channelP
+                   and *channelInfoP as it terminates.
+                */
             }
         } else {
             /* Accept function was interrupted before it got a connection */
+            trace(&srvP->tracer,
+                  "Wait for new channel from switch was interrupted");
+            *errorP = NULL;
         }
     }
 }
@@ -1079,19 +1290,37 @@ acceptAndProcessNextConnection(
 
 
 static void 
-serverRun2(TServer * const serverP) {
+serverRun2(TServer *     const serverP,
+           const char ** const errorP) {
 
     struct _TServer * const srvP = serverP->srvP;
     outstandingConnList * outstandingConnListP;
 
     createOutstandingConnList(&outstandingConnListP);
 
-    while (!srvP->terminationRequested)
-        acceptAndProcessNextConnection(serverP, outstandingConnListP);
+    *errorP = NULL;  /* initial value */
 
-    waitForNoConnections(outstandingConnListP);
+    trace(&srvP->tracer, "Starting main connection accepting loop");
     
-    destroyOutstandingConnList(outstandingConnListP);
+    while (!srvP->terminationRequested && !*errorP)
+        acceptAndProcessNextConnection(serverP, outstandingConnListP, errorP);
+
+    trace(&srvP->tracer, "Main connection accepting loop is done");
+
+    if (!*errorP) {
+        trace(&srvP->tracer,
+              "Interrupting and waiting for %u existing connections "
+              "to finish",
+              outstandingConnListP->count);
+
+        interruptChannels(outstandingConnListP);
+
+        waitForNoConnections(outstandingConnListP);
+    
+        trace(&srvP->tracer, "No connections left");
+
+        destroyOutstandingConnList(outstandingConnListP);
+    }
 }
 
 
@@ -1101,12 +1330,27 @@ ServerRun(TServer * const serverP) {
 
     struct _TServer * const srvP = serverP->srvP;
 
-    if (!srvP->chanSwitchP)
+    trace(&srvP->tracer, "%s entered", __FUNCTION__);
+
+    if (!srvP->serverAcceptsConnections)
         TraceMsg("This server is not set up to accept connections "
                  "on its own, so you can't use ServerRun().  "
-                 "Try ServerRunConn() or ServerInit()");
-    else
-        serverRun2(serverP);
+                 "Try ServerRunConn()");
+    else if (!srvP->readyToAccept)
+        TraceMsg("You cannot run this server until you initialize it to "
+                 "accept connections, with ServerInit()");
+    else {
+        const char * error;
+
+        serverRun2(serverP, &error);
+
+        if (error) {
+            TraceMsg("Server failed.  %s", error);
+
+            xmlrpc_strfree(error);
+        }
+    }
+    trace(&srvP->tracer, "%s exiting", __FUNCTION__);
 }
 
 
@@ -1128,6 +1372,8 @@ serverRunChannel(TServer *     const serverP,
     TConn * connectionP;
     const char * error;
 
+    trace(&srvP->tracer, "%s entered", __FUNCTION__);
+
     srvP->keepalivemaxconn = 1;
 
     ConnCreate(&connectionP, 
@@ -1146,6 +1392,7 @@ serverRunChannel(TServer *     const serverP,
 
         ConnWaitAndRelease(connectionP);
     }
+    trace(&srvP->tracer, "%s exiting", __FUNCTION__);
 }
 
 
@@ -1163,6 +1410,8 @@ ServerRunChannel(TServer *     const serverP,
 -----------------------------------------------------------------------------*/
     struct _TServer * const srvP = serverP->srvP;
 
+    trace(&srvP->tracer, "%s entered", __FUNCTION__);
+
     if (srvP->serverAcceptsConnections)
         xmlrpc_asprintf(errorP,
                         "This server is configured to accept connections on "
@@ -1170,6 +1419,8 @@ ServerRunChannel(TServer *     const serverP,
                         "Try ServerRun() or ServerCreateNoAccept().");
     else
         serverRunChannel(serverP, channelP, channelInfoP, errorP);
+
+    trace(&srvP->tracer, "%s exiting", __FUNCTION__);
 }
 
 
@@ -1231,18 +1482,23 @@ ServerRunConn(TServer * const serverP,
 void
 ServerRunOnce(TServer * const serverP) {
 /*----------------------------------------------------------------------------
-   Accept a connection from the channel switch and do the HTTP
-   transaction that comes over it.
+   Accept a connection from the channel switch and do the HTTP transaction
+   that comes over it.
 
-   If no connection is presently waiting at the switch, wait for one.
-   But return immediately if we receive a signal during the wait.
+   If no connection is presently waiting at the switch, wait for one.  But
+   return immediately if we receive a signal during the wait.
 -----------------------------------------------------------------------------*/
     struct _TServer * const srvP = serverP->srvP;
 
-    if (!srvP->chanSwitchP)
+    trace(&srvP->tracer, "%s entered", __FUNCTION__);
+
+    if (!srvP->serverAcceptsConnections)
         TraceMsg("This server is not set up to accept connections "
-                 "on its own, so you can't use ServerRunOnce().  "
-                 "Try ServerRunChannel() or ServerInit()");
+                 "on its own, so you can't use ServerRun().  "
+                 "Try ServerRunConn()");
+    else if (!srvP->readyToAccept)
+        TraceMsg("You cannot run this server until you initialize it to "
+                 "accept connections, with ServerInit()");
     else {
         const char * error;
         TChannel *   channelP;
@@ -1250,6 +1506,8 @@ ServerRunOnce(TServer * const serverP) {
     
         srvP->keepalivemaxconn = 1;
 
+        assert(srvP->chanSwitchP);
+
         ChanSwitchAccept(srvP->chanSwitchP, &channelP, &channelInfoP, &error);
         if (error) {
             TraceMsg("Failed to accept the next connection from a client "
@@ -1274,6 +1532,7 @@ ServerRunOnce(TServer * const serverP) {
             }
         }
     }
+    trace(&srvP->tracer, "%s exiting", __FUNCTION__);
 }
 
 
@@ -1297,11 +1556,14 @@ ServerRunOnce2(TServer *           const serverP,
 
 
 static void
-setGroups(void) {
+setGroups(const char ** const errorP) {
 
 #if HAVE_SETGROUPS   
-    if (setgroups(0, NULL) == (-1))
-        TraceExit("Failed to setup the group.");
+    if (setgroups(0, NULL) == -1)
+        xmlrpc_asprintf(errorP, "setgroups() errno = %d (%s)",
+                        errno, strerror(errno));
+#else
+    *errorP = NULL;
 #endif
 }
 
@@ -1322,7 +1584,7 @@ ServerDaemonize(TServer * const serverP) {
 -----------------------------------------------------------------------------*/
     struct _TServer * const srvP = serverP->srvP;
 
-#ifndef _WIN32
+#if !MSVCRT
     /* Become a daemon */
     switch (fork()) {
     case 0:
@@ -1338,12 +1600,18 @@ ServerDaemonize(TServer * const serverP) {
 
     /* Change the current user if we are root */
     if (getuid()==0) {
+        const char * error;
         if (srvP->uid == (uid_t)-1)
             TraceExit("Can't run under root privileges.  "
                       "Please add a User option in your "
                       "Abyss configuration file.");
 
-        setGroups();
+        setGroups(&error);
+
+        if (error) {
+            TraceExit("Failed to set groups.  %s", error);
+            xmlrpc_strfree(error);
+        }
 
         if (srvP->gid != (gid_t)-1)
             if (setgid(srvP->gid)==(-1))
@@ -1356,11 +1624,11 @@ ServerDaemonize(TServer * const serverP) {
     if (srvP->pidfileP) {
         char z[16];
     
-        sprintf(z, "%d", getpid());
+        sprintf(z, "%d", XMLRPC_GETPID());
         FileWrite(srvP->pidfileP, z, strlen(z));
         FileClose(srvP->pidfileP);
     }
-#endif  /* _WIN32 */
+#endif  /* MSVCRT */
 }
 
 
@@ -1384,7 +1652,7 @@ serverAddHandler(TServer *     const serverP,
 
     MALLOCVAR(handlerP);
     if (handlerP == NULL)
-        *successP = FALSE;
+        *successP = false;
     else {
         handlerP->init       = init;
         handlerP->term       = term;
@@ -1397,7 +1665,7 @@ serverAddHandler(TServer *     const serverP,
             MAX(srvP->uriHandlerStackSize, handleReqStackSize);
         
         if (handlerP->init == NULL)
-            *successP = TRUE;
+            *successP = true;
         else {
             URIHandler2 handler2 = makeUriHandler2(handlerP);
             handlerP->init(&handler2, successP);
@@ -1506,20 +1774,24 @@ LogWrite(TServer *    const serverP,
 
     struct _TServer * const srvP = serverP->srvP;
 
-    if (!srvP->logfileisopen && srvP->logfilename)
-        logOpen(srvP);
+    if (!srvP->logfileisopen && srvP->logfilename) {
+        const char * error;
+        logOpen(srvP, &error);
 
-    if (srvP->logfileisopen) {
-        bool success;
-        success = MutexLock(srvP->logmutexP);
-        if (success) {
-            const char * const lbr = "\n";
-            FileWrite(srvP->logfileP, msg, strlen(msg));
-            FileWrite(srvP->logfileP, lbr, strlen(lbr));
-        
-            MutexUnlock(srvP->logmutexP);
+        if (error) {
+            TraceMsg("Failed to open log file.  %s", error);
+
+            xmlrpc_strfree(error);
         }
     }
+    if (srvP->logfileisopen) {
+        const char * const lbr = "\n";
+        srvP->logLockP->acquire(srvP->logLockP);
+        FileWrite(srvP->logfileP, msg, strlen(msg));
+        FileWrite(srvP->logfileP, lbr, strlen(lbr));
+        
+        srvP->logLockP->release(srvP->logLockP);
+    }
 }
 /*******************************************************************************
 **
