@@ -20,7 +20,7 @@ using girerr::throwf;
 #include "xmlrpc-c/girmem.hpp"
 using girmem::autoObjectPtr;
 using girmem::autoObject;
-#include "env_wrap.hpp"
+#include "xmlrpc-c/env_wrap.hpp"
 #include "xmlrpc-c/base.h"
 #include "xmlrpc-c/client.h"
 #include "xmlrpc-c/transport.h"
@@ -133,10 +133,11 @@ struct clientXmlTransport_curl::constrOpt_impl {
 
     struct {
         std::string  network_interface;
-        bool         no_ssl_verifypeer;
-        bool         no_ssl_verifyhost;
         bool         dont_advertise;
         std::string  user_agent;
+        std::string  referer;
+        bool         no_ssl_verifypeer;
+        bool         no_ssl_verifyhost;
         std::string  ssl_cert;
         std::string  sslcerttype;
         std::string  sslcertpasswd;
@@ -157,13 +158,16 @@ struct clientXmlTransport_curl::constrOpt_impl {
         unsigned int proxy_port;
         std::string  proxy_userpwd;
         xmlrpc_httpproxytype proxy_type;
+        bool         gssapi_delegation;
+        unsigned int connect_timeout;
     } value;
     struct {
         bool network_interface;
-        bool no_ssl_verifypeer;
-        bool no_ssl_verifyhost;
         bool dont_advertise;
         bool user_agent;
+        bool referer;
+        bool no_ssl_verifypeer;
+        bool no_ssl_verifyhost;
         bool ssl_cert;
         bool sslcerttype;
         bool sslcertpasswd;
@@ -184,16 +188,19 @@ struct clientXmlTransport_curl::constrOpt_impl {
         bool proxy_port;
         bool proxy_userpwd;
         bool proxy_type;
+        bool gssapi_delegation;
+        bool connect_timeout;
     } present;
 };
 
 clientXmlTransport_curl::constrOpt_impl::constrOpt_impl() {
 
     present.network_interface = false;
-    present.no_ssl_verifypeer = false;
-    present.no_ssl_verifyhost = false;
     present.dont_advertise    = false;
     present.user_agent        = false;
+    present.referer           = false;
+    present.no_ssl_verifypeer = false;
+    present.no_ssl_verifyhost = false;
     present.ssl_cert          = false;
     present.sslcerttype       = false;
     present.sslcertpasswd     = false;
@@ -214,6 +221,8 @@ clientXmlTransport_curl::constrOpt_impl::constrOpt_impl() {
     present.proxy_auth        = false;
     present.proxy_userpwd     = false;
     present.proxy_type        = false;
+    present.gssapi_delegation = false;
+    present.connect_timeout   = false;
 }
 
 
@@ -227,10 +236,11 @@ clientXmlTransport_curl::constrOpt::OPTION_NAME(TYPE const& arg) { \
 }
 
 DEFINE_OPTION_SETTER(network_interface, string);
+DEFINE_OPTION_SETTER(dont_advertise, bool);
+DEFINE_OPTION_SETTER(referer, string);
+DEFINE_OPTION_SETTER(user_agent, string);
 DEFINE_OPTION_SETTER(no_ssl_verifypeer, bool);
 DEFINE_OPTION_SETTER(no_ssl_verifyhost, bool);
-DEFINE_OPTION_SETTER(dont_advertise, bool);
-DEFINE_OPTION_SETTER(user_agent, string);
 DEFINE_OPTION_SETTER(ssl_cert, string);
 DEFINE_OPTION_SETTER(sslcerttype, string);
 DEFINE_OPTION_SETTER(sslcertpasswd, string);
@@ -251,6 +261,8 @@ DEFINE_OPTION_SETTER(proxy_port, unsigned int);
 DEFINE_OPTION_SETTER(proxy_auth, unsigned int);
 DEFINE_OPTION_SETTER(proxy_userpwd, string);
 DEFINE_OPTION_SETTER(proxy_type, xmlrpc_httpproxytype);
+DEFINE_OPTION_SETTER(gssapi_delegation, bool);
+DEFINE_OPTION_SETTER(connect_timeout, unsigned int);
 
 #undef DEFINE_OPTION_SETTER
 
@@ -285,14 +297,16 @@ clientXmlTransport_curl::initialize(constrOpt const& optExt) {
 
     transportParms.network_interface = opt.present.network_interface ?
         opt.value.network_interface.c_str() : NULL;
-    transportParms.no_ssl_verifypeer = opt.present.no_ssl_verifypeer ? 
-        opt.value.no_ssl_verifypeer         : false;
-    transportParms.no_ssl_verifyhost = opt.present.no_ssl_verifyhost ? 
-        opt.value.no_ssl_verifyhost         : false;
+    transportParms.referer           = opt.present.referer ? 
+        opt.value.referer.c_str()     : NULL;
     transportParms.dont_advertise    = opt.present.dont_advertise ?
         opt.value.dont_advertise            : false;
     transportParms.user_agent        = opt.present.user_agent ?
         opt.value.user_agent.c_str()        : NULL;
+    transportParms.no_ssl_verifypeer = opt.present.no_ssl_verifypeer ? 
+        opt.value.no_ssl_verifypeer         : false;
+    transportParms.no_ssl_verifyhost = opt.present.no_ssl_verifyhost ? 
+        opt.value.no_ssl_verifyhost         : false;
     transportParms.ssl_cert          = opt.present.ssl_cert ?
         opt.value.ssl_cert.c_str()          : NULL;
     transportParms.sslcerttype       = opt.present.sslcerttype ?
@@ -333,6 +347,10 @@ clientXmlTransport_curl::initialize(constrOpt const& optExt) {
         opt.value.proxy_userpwd.c_str()     : NULL;
     transportParms.proxy_type        = opt.present.proxy_type ? 
         opt.value.proxy_type                : XMLRPC_HTTPPROXY_HTTP;
+    transportParms.gssapi_delegation = opt.present.gssapi_delegation ?
+        opt.value.gssapi_delegation         : false;
+    transportParms.connect_timeout   = opt.present.connect_timeout ? 
+        opt.value.connect_timeout           : 0;
 
     this->c_transportOpsP = &xmlrpc_curl_transport_ops;
 
@@ -340,7 +358,7 @@ clientXmlTransport_curl::initialize(constrOpt const& optExt) {
 
     xmlrpc_curl_transport_ops.create(
         &env.env_c, 0, "", "",
-        &transportParms, XMLRPC_CXPSIZE(proxy_userpwd),
+        &transportParms, XMLRPC_CXPSIZE(gssapi_delegation),
         &this->c_transportP);
 
     if (env.env_c.fault_occurred)
