@@ -3,6 +3,7 @@
 #include "xmlrpc_config.h"
 
 #define _XOPEN_SOURCE 600  /* For strdup(), sigaction */
+#define WIN32_LEAN_AND_MEAN  /* required by xmlrpc-c/abyss.h */
 
 #include <assert.h>
 #include <stdio.h>
@@ -11,25 +12,29 @@
 #include <errno.h>
 #include <time.h>
 #include <fcntl.h>
-#ifdef _WIN32
+#if MSVCRT
 #  include <io.h>
+#  pragma comment(lib, "Ws2_32.lib")
 #else
 #  include <signal.h>
 #  include <sys/wait.h>
 #  include <grp.h>
+#  include <sys/socket.h>
+#  include <netinet/in.h>
 #endif
 
 #include "bool.h"
 #include "int.h"
 #include "mallocvar.h"
-#include "xmlrpc-c/abyss.h"
 
+#include "xmlrpc-c/abyss.h"
 #include "xmlrpc-c/base.h"
 #include "xmlrpc-c/server.h"
 #include "xmlrpc-c/base_int.h"
 #include "xmlrpc-c/string_int.h"
 #include "xmlrpc-c/server_abyss.h"
 
+#include "abyss_handler.h"
 
 
 struct xmlrpc_server_abyss {
@@ -123,375 +128,6 @@ validateGlobalInit(xmlrpc_env * const envP) {
 
 
 
-static void
-addAuthCookie(xmlrpc_env * const envP,
-              TSession *   const abyssSessionP,
-              const char * const authCookie) {
-
-    const char * cookieResponse;
-    
-    xmlrpc_asprintf(&cookieResponse, "auth=%s", authCookie);
-    
-    if (xmlrpc_strnomem(cookieResponse))
-        xmlrpc_faultf(envP, "Insufficient memory to generate cookie "
-                      "response header.");
-    else {
-        ResponseAddField(abyssSessionP, "Set-Cookie", cookieResponse);
-    
-        xmlrpc_strfree(cookieResponse);
-    }
-}   
-    
-
-
-static void 
-sendResponse(xmlrpc_env *      const envP,
-             TSession *        const abyssSessionP, 
-             const char *      const body, 
-             size_t            const len,
-             bool              const chunked,
-             ResponseAccessCtl const accessControl) {
-/*----------------------------------------------------------------------------
-   Generate an HTTP response containing body 'body' of length 'len'
-   characters.
-
-   This is meant to run in the context of an Abyss URI handler for
-   Abyss session 'abyssSessionP'.
------------------------------------------------------------------------------*/
-    const char * http_cookie = NULL;
-        /* This used to set http_cookie to getenv("HTTP_COOKIE"), but
-           that doesn't make any sense -- environment variables are not
-           appropriate for this.  So for now, cookie code is disabled.
-           - Bryan 2004.10.03.
-        */
-
-    /* Various bugs before Xmlrpc-c 1.05 caused the response to be not
-       chunked in the most basic case, but chunked if the client explicitly
-       requested keepalive.  I think it's better not to chunk, because
-       it's simpler, so I removed this in 1.05.  I don't know what the
-       purpose of chunking would be, and an original comment suggests
-       the author wasn't sure chunking was a good idea.
-
-       In 1.06 we added the user option to chunk.
-    */
-    if (chunked)
-        ResponseChunked(abyssSessionP);
-
-    ResponseStatus(abyssSessionP, 200);
-
-    if (http_cookie)
-        /* There's an auth cookie, so pass it back in the response. */
-        addAuthCookie(envP, abyssSessionP, http_cookie);
-
-    if ((size_t)(uint32_t)len != len)
-        xmlrpc_faultf(envP, "XML-RPC method generated a response too "
-                      "large for Abyss to send");
-    else {
-        uint32_t const abyssLen = (uint32_t)len;
-
-        /* See discussion below of quotes around "utf-8" */
-        ResponseContentType(abyssSessionP, "text/xml; charset=utf-8");
-        ResponseContentLength(abyssSessionP, abyssLen);
-        ResponseAccessControl(abyssSessionP, accessControl);
-        
-        ResponseWriteStart(abyssSessionP);
-        ResponseWriteBody(abyssSessionP, body, abyssLen);
-        ResponseWriteEnd(abyssSessionP);
-    }
-}
-
-
-
-/* From 0.9.10 (May 2001) through 1.17 (December 2008), the content-type
-   header said charset="utf-8" (i.e. with the value of 'charset' an HTTP quoted
-   string).  Before 0.9.10, the header didn't have charset at all.
-
-   We got a complaint in January 2009 that some client didn't understand that,
-   saying
-
-     apache2: XML-RPC: xmlrpcmsg::parseResponse: invalid charset encoding of
-     received response: "UTF-8"
-
-   And that removing the quotation marks fixes this.
-
-   From what I can tell, the module is wrong to distinguish between the
-   two, but I don't think it hurts anything to use a basic HTTP token instead
-   of an HTTP quoted string here, so starting in 1.18, we do.  */
-
-
-static void
-sendError(TSession *   const abyssSessionP, 
-          unsigned int const status,
-          const char * const explanation) {
-/*----------------------------------------------------------------------------
-  Send an error response back to the client.
------------------------------------------------------------------------------*/
-    ResponseStatus(abyssSessionP, (uint16_t) status);
-    ResponseError2(abyssSessionP, explanation);
-}
-
-
-
-static void
-traceChunkRead(TSession * const abyssSessionP) {
-
-    fprintf(stderr, "XML-RPC handler got a chunk of %u bytes\n",
-            (unsigned int)SessionReadDataAvail(abyssSessionP));
-}
-
-
-
-static void
-refillBufferFromConnection(xmlrpc_env * const envP,
-                           TSession *   const abyssSessionP,
-                           const char * const trace) {
-/*----------------------------------------------------------------------------
-   Get the next chunk of data from the connection into the buffer.
------------------------------------------------------------------------------*/
-    bool succeeded;
-
-    succeeded = SessionRefillBuffer(abyssSessionP);
-
-    if (!succeeded)
-        xmlrpc_env_set_fault_formatted(
-            envP, XMLRPC_TIMEOUT_ERROR, "Timed out waiting for "
-            "client to send its POST data");
-    else {
-        if (trace)
-            traceChunkRead(abyssSessionP);
-    }
-}
-
-
-
-static void
-getBody(xmlrpc_env *        const envP,
-        TSession *          const abyssSessionP,
-        size_t              const contentSize,
-        const char *        const trace,
-        xmlrpc_mem_block ** const bodyP) {
-/*----------------------------------------------------------------------------
-   Get the entire body, which is of size 'contentSize' bytes, from the
-   Abyss session and return it as the new memblock *bodyP.
-
-   The first chunk of the body may already be in Abyss's buffer.  We
-   retrieve that before reading more.
------------------------------------------------------------------------------*/
-    xmlrpc_mem_block * body;
-
-    if (trace)
-        fprintf(stderr, "XML-RPC handler processing body.  "
-                "Content Size = %u bytes\n", (unsigned)contentSize);
-
-    body = xmlrpc_mem_block_new(envP, 0);
-    if (!envP->fault_occurred) {
-        size_t bytesRead;
-        const char * chunkPtr;
-        size_t chunkLen;
-
-        bytesRead = 0;
-
-        while (!envP->fault_occurred && bytesRead < contentSize) {
-            SessionGetReadData(abyssSessionP, contentSize - bytesRead, 
-                               &chunkPtr, &chunkLen);
-            bytesRead += chunkLen;
-
-            assert(bytesRead <= contentSize);
-
-            XMLRPC_MEMBLOCK_APPEND(char, envP, body, chunkPtr, chunkLen);
-            if (bytesRead < contentSize)
-                refillBufferFromConnection(envP, abyssSessionP, trace);
-        }
-        if (envP->fault_occurred)
-            xmlrpc_mem_block_free(body);
-    }
-    *bodyP = body;
-}
-
-
-
-static void
-storeCookies(TSession *     const httpRequestP,
-             const char **  const errorP) {
-/*----------------------------------------------------------------------------
-   Get the cookie settings from the HTTP headers and remember them for
-   use in responses.
------------------------------------------------------------------------------*/
-    const char * const cookie = RequestHeaderValue(httpRequestP, "cookie");
-    if (cookie) {
-        /* 
-           Setting the value in an environment variable doesn't make
-           any sense.  So for now, cookie code is disabled.
-           -Bryan 04.10.03.
-
-        setenv("HTTP_COOKIE", cookie, 1);
-        */
-    }
-    /* TODO: parse HTTP_COOKIE to find auth pair, if there is one */
-
-    *errorP = NULL;
-}
-
-
-
-
-static void
-processContentLength(TSession *    const httpRequestP,
-                     size_t *      const inputLenP,
-                     bool *        const missingP,
-                     const char ** const errorP) {
-/*----------------------------------------------------------------------------
-  Make sure the content length is present and non-zero.  This is
-  technically required by XML-RPC, but we only enforce it because we
-  don't want to figure out how to safely handle HTTP < 1.1 requests
-  without it.
------------------------------------------------------------------------------*/
-    const char * const content_length = 
-        RequestHeaderValue(httpRequestP, "content-length");
-
-    if (content_length == NULL) {
-        *missingP = TRUE;
-        *errorP = NULL;
-    } else {
-        *missingP = FALSE;
-        *inputLenP = 0;  /* quiet compiler warning */
-        if (content_length[0] == '\0')
-            xmlrpc_asprintf(errorP, "The value in your content-length "
-                            "HTTP header value is a null string");
-        else {
-            unsigned long contentLengthValue;
-            char * tail;
-        
-            contentLengthValue = strtoul(content_length, &tail, 10);
-        
-            if (*tail != '\0')
-                xmlrpc_asprintf(errorP, "There's non-numeric crap in "
-                                "the value of your content-length "
-                                "HTTP header: '%s'", tail);
-            else if (contentLengthValue < 1)
-                xmlrpc_asprintf(errorP, "According to your content-length "
-                                "HTTP header, your request is empty (zero "
-                                "length)");
-            else if ((unsigned long)(size_t)contentLengthValue 
-                     != contentLengthValue)
-                xmlrpc_asprintf(errorP, "According to your content-length "
-                                "HTTP header, your request is too big to "
-                                "process; we can't even do arithmetic on its "
-                                "size: %s bytes", content_length);
-            else {
-                *errorP = NULL;
-                *inputLenP = (size_t)contentLengthValue;
-            }
-        }
-    }
-}
-
-
-
-static void
-traceHandlerCalled(TSession * const abyssSessionP) {
-    
-    const char * methodDesc;
-    const TRequestInfo * requestInfoP;
-
-    fprintf(stderr, "xmlrpc_server_abyss URI path handler called.\n");
-
-    SessionGetRequestInfo(abyssSessionP, &requestInfoP);
-
-    fprintf(stderr, "URI = '%s'\n", requestInfoP->uri);
-
-    switch (requestInfoP->method) {
-    case m_unknown: methodDesc = "unknown";   break;
-    case m_get:     methodDesc = "get";       break;
-    case m_put:     methodDesc = "put";       break;
-    case m_head:    methodDesc = "head";      break;
-    case m_post:    methodDesc = "post";      break;
-    case m_delete:  methodDesc = "delete";    break;
-    case m_trace:   methodDesc = "trace";     break;
-    case m_options: methodDesc = "m_options"; break;
-    default:        methodDesc = "?";
-    }
-    fprintf(stderr, "HTTP method = '%s'\n", methodDesc);
-
-    if (requestInfoP->query)
-        fprintf(stderr, "query (component of URL)='%s'\n",
-                requestInfoP->query);
-    else
-        fprintf(stderr, "URL has no query component\n");
-}
-
-
-
-static void
-processCall(TSession *            const abyssSessionP,
-            size_t                const contentSize,
-            xmlrpc_call_processor       xmlProcessor,
-            void *                const xmlProcessorArg,
-            bool                  const wantChunk,
-            ResponseAccessCtl     const accessControl,
-            const char *          const trace) {
-/*----------------------------------------------------------------------------
-   Handle an RPC request.  This is an HTTP request that has the proper form
-   to be an XML-RPC call.
-
-   The text of the call is available through the Abyss session
-   'abyssSessionP'.
-
-   Its content length is 'contentSize' bytes.
------------------------------------------------------------------------------*/
-    xmlrpc_env env;
-
-    if (trace)
-        fprintf(stderr,
-                "xmlrpc_server_abyss URI path handler processing RPC.\n");
-
-    xmlrpc_env_init(&env);
-
-    if (contentSize > xmlrpc_limit_get(XMLRPC_XML_SIZE_LIMIT_ID))
-        xmlrpc_env_set_fault_formatted(
-            &env, XMLRPC_LIMIT_EXCEEDED_ERROR,
-            "XML-RPC request too large (%u bytes)", (unsigned)contentSize);
-    else {
-        xmlrpc_mem_block * body;
-        /* Read XML data off the wire. */
-        getBody(&env, abyssSessionP, contentSize, trace, &body);
-        if (!env.fault_occurred) {
-            xmlrpc_mem_block * output;
-
-            /* Process the RPC. */
-            xmlProcessor(
-                &env, xmlProcessorArg,
-                XMLRPC_MEMBLOCK_CONTENTS(char, body),
-                XMLRPC_MEMBLOCK_SIZE(char, body),
-                abyssSessionP,
-                &output);
-            if (!env.fault_occurred) {
-                /* Send out the result. */
-                sendResponse(&env, abyssSessionP, 
-                             XMLRPC_MEMBLOCK_CONTENTS(char, output),
-                             XMLRPC_MEMBLOCK_SIZE(char, output),
-                             wantChunk, accessControl);
-                
-                XMLRPC_MEMBLOCK_FREE(char, output);
-            }
-            XMLRPC_MEMBLOCK_FREE(char, body);
-        }
-    }
-    if (env.fault_occurred) {
-        uint16_t httpResponseStatus;
-        if (env.fault_code == XMLRPC_TIMEOUT_ERROR)
-            httpResponseStatus = 408;  /* Request Timeout */
-        else
-            httpResponseStatus = 500;  /* Internal Server Error */
-
-        sendError(abyssSessionP, httpResponseStatus, env.fault_string);
-    }
-
-    xmlrpc_env_clean(&env);
-}
-
-
-
 static void
 processXmlrpcCall(xmlrpc_env *        const envP,
                   void *              const arg,
@@ -510,216 +146,6 @@ processXmlrpcCall(xmlrpc_env *        const envP,
 
 
 
-static const char * trace_abyss;
-
-
-struct uriHandlerXmlrpc {
-/*----------------------------------------------------------------------------
-   This is the part of an Abyss HTTP request handler (aka URI handler)
-   that is specific to the Xmlrpc-c handler.
------------------------------------------------------------------------------*/
-    xmlrpc_registry *       registryP;
-    const char *            uriPath;  /* malloc'ed */
-    bool                    chunkResponse;
-        /* The handler should chunk its response whenever possible */
-    xmlrpc_call_processor * xmlProcessor;
-    void *                  xmlProcessorArg;
-    ResponseAccessCtl       accessControl;
-};
-
-
-
-static void
-termAccessControl(ResponseAccessCtl * const accessCtlP) {
-
-    xmlrpc_strfreenull(accessCtlP->allowOrigin);
-}
-
-
-
-static void
-termUriHandler(void * const arg) {
-
-    struct uriHandlerXmlrpc * const uriHandlerXmlrpcP = arg;
-
-    xmlrpc_strfree(uriHandlerXmlrpcP->uriPath);
-    termAccessControl(&uriHandlerXmlrpcP->accessControl);
-    free(uriHandlerXmlrpcP);
-}
-
-
-
-static void
-handleXmlRpcCallReq(TSession *           const abyssSessionP,
-                    const TRequestInfo * const requestInfoP ATTR_UNUSED,
-                    xmlrpc_call_processor      xmlProcessor,
-                    void *               const xmlProcessorArg,
-                    bool                 const wantChunk,
-                    ResponseAccessCtl    const accessControl) {
-/*----------------------------------------------------------------------------
-   Handle the HTTP request described by *requestInfoP, which arrived over
-   Abyss HTTP session *abyssSessionP, which is an XML-RPC call
-   (i.e. a POST request to /RPC2 or whatever other URI our server is
-   supposed to handle).
-
-   Handle it by feeding the XML which is its content to 'xmlProcessor'
-   along with argument 'xmlProcessorArg'.
------------------------------------------------------------------------------*/
-    /* We used to reject the call if content-type was not present and
-       text/xml, on some security theory (a firewall may block text/xml with
-       the intent of blocking XML-RPC.  Now, we believe that is silly, and we
-       have seen an incorrectly implemented client that says text/plain.
-    */
-    const char * error;
-
-    assert(requestInfoP->method == m_post);
-
-    storeCookies(abyssSessionP, &error);
-    if (error) {
-        sendError(abyssSessionP, 400, error);
-        xmlrpc_strfree(error);
-    } else {
-        const char * error;
-        bool missing;
-        size_t contentSize;
-
-        processContentLength(abyssSessionP, 
-                             &contentSize, &missing, &error);
-        if (error) {
-            sendError(abyssSessionP, 400, error);
-            xmlrpc_strfree(error);
-        } else {
-            if (missing)
-                sendError(abyssSessionP, 411, "You must send a "
-                          "content-length HTTP header in an "
-                          "XML-RPC call.");
-            else
-                processCall(abyssSessionP, contentSize,
-                            xmlProcessor, xmlProcessorArg,
-                            wantChunk, accessControl,
-                            trace_abyss);
-        }
-    }
-}
-
-
-
-static void
-handleXmlRpcOptionsReq(TSession *        const abyssSessionP,
-                       ResponseAccessCtl const accessControl) {
-
-    ResponseAddField(abyssSessionP, "Allow", "POST");
-    
-    ResponseAccessControl(abyssSessionP, accessControl);
-    ResponseContentLength(abyssSessionP, 0);
-    ResponseStatus(abyssSessionP, 200);
-    ResponseWriteStart(abyssSessionP);
-    ResponseWriteEnd(abyssSessionP);
-}
-
-
-
-static void
-handleIfXmlrpcReq(void *        const handlerArg,
-                  TSession *    const abyssSessionP,
-                  abyss_bool *  const handledP) {
-/*----------------------------------------------------------------------------
-   Our job is to look at this HTTP request that the Abyss server is
-   trying to process and see if we can handle it.  If it's an XML-RPC
-   call for this XML-RPC server, we handle it.  If it's not, we refuse
-   it and Abyss can try some other handler.
-
-   Our return code is TRUE to mean we handled it; FALSE to mean we didn't.
-
-   Note that failing the request counts as handling it, and not handling
-   it does not mean we failed it.
-
-   This is an Abyss HTTP Request handler -- type handleReqFn3.
------------------------------------------------------------------------------*/
-    struct uriHandlerXmlrpc * const uriHandlerXmlrpcP = handlerArg;
-
-    const TRequestInfo * requestInfoP;
-
-    if (trace_abyss)
-        traceHandlerCalled(abyssSessionP);
-
-    SessionGetRequestInfo(abyssSessionP, &requestInfoP);
-
-    /* Note that requestInfoP->uri is not the whole URI.  It is just
-       the "file name" part of it.
-    */
-    if (!xmlrpc_streq(requestInfoP->uri, uriHandlerXmlrpcP->uriPath))
-        /* It's not for the path (e.g. "/RPC2") that we're supposed to
-           handle.
-        */
-        *handledP = FALSE;
-    else {
-        *handledP = TRUE;
-
-        switch (requestInfoP->method) {
-        case m_post:
-            handleXmlRpcCallReq(abyssSessionP, requestInfoP,
-                                uriHandlerXmlrpcP->xmlProcessor,
-                                uriHandlerXmlrpcP->xmlProcessorArg,
-                                uriHandlerXmlrpcP->chunkResponse,
-                                uriHandlerXmlrpcP->accessControl);
-            break;
-        case m_options:
-            handleXmlRpcOptionsReq(abyssSessionP,
-                                   uriHandlerXmlrpcP->accessControl);
-            break;
-        default:
-            sendError(abyssSessionP, 405,
-                      "POST is the only HTTP method this server understands");
-                /* 405 = Method Not Allowed */
-        }
-    }
-    if (trace_abyss)
-        fprintf(stderr, "xmlrpc_server_abyss URI path handler returning.\n");
-}
-
-
-/* This doesn't include what the user's method function requires */
-#define HANDLE_XMLRPC_REQ_STACK 1024
-
-
-
-/*=========================================================================
-**  xmlrpc_server_abyss_default_handler
-**=========================================================================
-**  This handler returns a 404 Not Found for all requests. See the header
-**  for more documentation.
-*/
-
-static xmlrpc_bool 
-xmlrpc_server_abyss_default_handler(TSession * const sessionP) {
-
-    const TRequestInfo * requestInfoP;
-
-    const char * explanation;
-
-    if (trace_abyss)
-        fprintf(stderr, "xmlrpc_server_abyss default handler called.\n");
-
-    SessionGetRequestInfo(sessionP, &requestInfoP);
-
-    xmlrpc_asprintf(
-        &explanation,
-        "This XML-RPC For C/C++ Abyss XML-RPC server "
-        "responds to only one URI path.  "
-        "I don't know what URI path that is, "
-        "but it's not the one you requested: '%s'.  (Typically, it's "
-        "'/RPC2')", requestInfoP->uri);
-
-    sendError(sessionP, 404, explanation);
-
-    xmlrpc_strfree(explanation);
-
-    return TRUE;
-}
-
-
-
 static void
 setHandler(xmlrpc_env *              const envP,
            TServer *                 const srvP,
@@ -728,14 +154,15 @@ setHandler(xmlrpc_env *              const envP,
     
     abyss_bool success;
 
-    trace_abyss = getenv("XMLRPC_TRACE_ABYSS");
+    xmlrpc_abyss_handler_trace(
+        getenv("XMLRPC_TRACE_ABYSS"));
                                  
     {
         size_t const stackSize = 
-            HANDLE_XMLRPC_REQ_STACK + xmlProcessorMaxStackSize;
+            xmlrpc_abyss_handler_stacksize() + xmlProcessorMaxStackSize;
         struct ServerReqHandler3 const handlerDesc = {
-            /* .term               = */ &termUriHandler,
-            /* .handleReq          = */ &handleIfXmlrpcReq,
+            /* .term               = */ &xmlrpc_termUriHandler,
+            /* .handleReq          = */ &xmlrpc_handleIfXmlrpcReq,
             /* .userdata           = */ uriHandlerXmlrpcP,
             /* .handleReqStackSize = */ stackSize
         };
@@ -754,21 +181,29 @@ interpretHttpAccessControl(
     unsigned int                              const parmSize,
     ResponseAccessCtl *                       const accessCtlP) {
 
+    const char * allowOrigin;
+    bool         expires;
+    unsigned int maxAge;
+
     if (parmSize >= XMLRPC_AHPSIZE(allow_origin) && parmsP->allow_origin)
-        accessCtlP->allowOrigin = xmlrpc_strdupsol(parmsP->allow_origin);
+        allowOrigin = parmsP->allow_origin;
     else
-        accessCtlP->allowOrigin = NULL;
+        allowOrigin = NULL;
 
     if (parmSize >= XMLRPC_AHPSIZE(access_ctl_expires)
         && parmsP->access_ctl_expires) {
-        accessCtlP->expires = true;
+        expires = true;
 
         if (parmSize >= XMLRPC_AHPSIZE(access_ctl_max_age))
-            accessCtlP->maxAge = parmsP->access_ctl_max_age;
+            maxAge = parmsP->access_ctl_max_age;
         else
-            accessCtlP->maxAge = 0;
-    } else
-        accessCtlP->expires = false;
+            maxAge = 0;
+    } else {
+        expires = false;
+        maxAge = 0;  /* Meaningless; just to quiet runtime memory checks */
+    }
+
+    xmlrpc_initAccessCtl(accessCtlP, allowOrigin, expires, maxAge);
 }
 
 
@@ -822,7 +257,7 @@ xmlrpc_server_abyss_set_handler3(
                                    &uriHandlerXmlrpcP->accessControl);
 
         if (envP->fault_occurred)
-            termAccessControl(&uriHandlerXmlrpcP->accessControl);
+            xmlrpc_termAccessControl(&uriHandlerXmlrpcP->accessControl);
     }
     if (!envP->fault_occurred)
         setHandler(envP, srvP, uriHandlerXmlrpcP, xmlProcessorMaxStackSize);
@@ -886,7 +321,7 @@ xmlrpc_server_abyss_set_handler(xmlrpc_env *      const envP,
 void
 xmlrpc_server_abyss_set_default_handler(TServer * const srvP) {
 
-    ServerDefaultHandler(srvP, xmlrpc_server_abyss_default_handler);
+    ServerDefaultHandler(srvP, xmlrpc_serverAbyssDefaultUriHandler);
 }
     
 
@@ -966,6 +401,79 @@ setAdditionalServerParms(const xmlrpc_server_abyss_parms * const parmsP,
         ServerSetTimeout(serverP, parmsP->timeout);
     if (parmSize >= XMLRPC_APSIZE(dont_advertise))
         ServerSetAdvertise(serverP, !parmsP->dont_advertise);
+    if (parmSize >= XMLRPC_APSIZE(max_conn)) {
+        if (parmsP->max_conn != 0)
+            ServerSetMaxConn(serverP, parmsP->max_conn);
+    }
+    if (parmSize >= XMLRPC_APSIZE(max_conn_backlog)) {
+        if (parmsP->max_conn_backlog != 0)
+            ServerSetMaxConnBacklog(serverP, parmsP->max_conn_backlog);
+    }
+    if (parmSize >= XMLRPC_APSIZE(max_rpc_mem)) {
+        if (parmsP->max_rpc_mem != 0)
+            ServerSetMaxSessionMem(serverP, parmsP->max_rpc_mem);
+    }
+}
+
+
+
+static void
+extractSockAddrParms(xmlrpc_env *                      const envP,
+                     const xmlrpc_server_abyss_parms * const parmsP,
+                     unsigned int                      const parmSize,
+                     const struct sockaddr **          const sockAddrPP,
+                     socklen_t *                       const sockAddrLenP,
+                     unsigned int *                    const portNumberP) {
+/*----------------------------------------------------------------------------
+   Return the server parameters that affect the address on which the server
+   socket shall listen.
+
+   There are two ways the arguments can specify this: 1) user supplies a
+   complete socket address, which specifies both a TCP port number and an IP
+   address (which determines on which interface, ergo which network, if any,
+   the server listens); and 2) just a TCP port number, which means he wants
+   to listen on all IPv4 interfaces and networks.  (2) is legacy.
+
+   If the user specifies the 'sockaddrP' and 'sockaddrlen' arguments, he gets
+   (1) and we ignore his 'port' argument.  We return his 'sockaddrP' and
+   'sockaddrlen' values as *sockAddrPP and *sockAddrLenP and nothing as
+   *portNumberP.
+
+   If the user doesn't specify 'sockaddrP', he gets (2).  We return NULL as
+   *sockAddrP and his 'port_number' argument as *portNumberP.  If he doesn't
+   specify 'port' either, we default it to 8080.
+
+   Specifying 'sockaddrP' and not 'sockaddrlen' is an error.
+
+   Note that the user's socket address may indicate "any IP address."
+-----------------------------------------------------------------------------*/
+    if (parmSize >= XMLRPC_APSIZE(sockaddr_p)) {
+        if (parmSize < XMLRPC_APSIZE(sockaddrlen))
+            xmlrpc_faultf(envP, "You must specify 'sockaddrlen' when you "
+                          "specify 'sockaddrP'");
+        else {
+            *sockAddrPP   = parmsP->sockaddr_p;
+            *sockAddrLenP = parmsP->sockaddrlen;
+        }
+    } else
+        *sockAddrPP = NULL;
+
+    if (*sockAddrPP == NULL) {
+        unsigned int portNumber;
+
+        if (parmSize >= XMLRPC_APSIZE(port_number))
+            portNumber = parmsP->port_number;
+        else
+            portNumber = 8080;
+        
+        if (portNumber > 0xffff)
+            xmlrpc_faultf(envP,
+                          "TCP port number %u exceeds the maximum possible "
+                          "TCP port number (65535)",
+                          portNumber);
+
+        *portNumberP = portNumber;
+    }        
 }
 
 
@@ -976,6 +484,8 @@ extractServerCreateParms(
     const xmlrpc_server_abyss_parms * const parmsP,
     unsigned int                      const parmSize,
     bool *                            const socketBoundP,
+    const struct sockaddr **          const sockAddrPP,
+    socklen_t *                       const sockAddrLenP,
     unsigned int *                    const portNumberP,
     TOsSocket *                       const socketFdP,
     const char **                     const logFileNameP) {
@@ -984,7 +494,7 @@ extractServerCreateParms(
     if (parmSize >= XMLRPC_APSIZE(socket_bound))
         *socketBoundP = parmsP->socket_bound;
     else
-        *socketBoundP = FALSE;
+        *socketBoundP = false;
 
     if (*socketBoundP) {
         if (parmSize < XMLRPC_APSIZE(socket_handle))
@@ -994,16 +504,8 @@ extractServerCreateParms(
         else
             *socketFdP = parmsP->socket_handle;
     } else {
-        if (parmSize >= XMLRPC_APSIZE(port_number))
-            *portNumberP = parmsP->port_number;
-        else
-            *portNumberP = 8080;
-
-        if (*portNumberP > 0xffff)
-            xmlrpc_faultf(envP,
-                          "TCP port number %u exceeds the maximum possible "
-                          "TCP port number (65535)",
-                          *portNumberP);
+        extractSockAddrParms(envP, parmsP, parmSize, sockAddrPP, sockAddrLenP,
+                             portNumberP);
     }
     if (!envP->fault_occurred) {
         if (parmSize >= XMLRPC_APSIZE(log_file_name) &&
@@ -1021,7 +523,7 @@ chanSwitchCreateOsSocket(TOsSocket      const socketFd,
                          TChanSwitch ** const chanSwitchPP,
                          const char **  const errorP) {
 
-#ifdef WIN32
+#if MSVCRT
     ChanSwitchWinCreateWinsock(socketFd, chanSwitchPP, errorP);
 #else
     ChanSwitchUnixCreateFd(socketFd, chanSwitchPP, errorP);
@@ -1032,35 +534,102 @@ chanSwitchCreateOsSocket(TOsSocket      const socketFd,
 
 
 static void
-createServerBoundSocket(xmlrpc_env *   const envP,
-                        TOsSocket      const socketFd,
-                        const char *   const logFileName,
-                        TServer *      const serverP,
-                        TChanSwitch ** const chanSwitchPP) {
+createChanSwitchOsSocket(xmlrpc_env *   const envP,
+                         TOsSocket      const socketFd,
+                         TChanSwitch ** const chanSwitchPP) {
 
-    TChanSwitch * chanSwitchP;
     const char * error;
-    
-    chanSwitchCreateOsSocket(socketFd, &chanSwitchP, &error);
+
+    chanSwitchCreateOsSocket(socketFd, chanSwitchPP, &error);
+
     if (error) {
-        xmlrpc_faultf(envP, "Unable to create Abyss socket out of "
+        xmlrpc_faultf(envP, "Unable to create Abyss channel switch out of "
                       "file descriptor %d.  %s", socketFd, error);
         xmlrpc_strfree(error);
-    } else {
-        ServerCreateSwitch(serverP, chanSwitchP, &error);
+    }
+}
+
+
+
+static void
+chanSwitchCreateSockAddr(int                     const protocolFamily,
+                         const struct sockaddr * const sockAddrP,
+                         socklen_t               const sockAddrLen,
+                         TChanSwitch **          const chanSwitchPP,
+                         const char **           const errorP) {
+
+#if MSVCRT
+    ChanSwitchWinCreate2(protocolFamily, sockAddrP, sockAddrLen, 
+                          chanSwitchPP, errorP);
+#else
+    ChanSwitchUnixCreate2(protocolFamily, sockAddrP, sockAddrLen, 
+                          chanSwitchPP, errorP);
+#endif
+
+}
+
+
+
+static void
+createChanSwitchSockAddr(xmlrpc_env *            const envP,
+                         const struct sockaddr * const sockAddrP,
+                         socklen_t               const sockAddrLen,
+                         TChanSwitch **          const chanSwitchPP) {
+
+    int protocolFamily;
+
+    assert(sockAddrP);
+
+    switch (sockAddrP->sa_family) {
+    case AF_INET:
+        protocolFamily = PF_INET;
+        break;
+    case AF_INET6:
+        protocolFamily = PF_INET6;
+        break;
+    default:
+        xmlrpc_faultf(envP, "Unknown socket address family %d.  "
+                      "We know only AF_INET and AF_INET6.",
+                      sockAddrP->sa_family);
+    }
+
+    if (!envP->fault_occurred) {
+        const char * error;
+
+        chanSwitchCreateSockAddr(protocolFamily, sockAddrP, sockAddrLen,
+                                 chanSwitchPP, &error);
+
         if (error) {
-            xmlrpc_faultf(envP, "Abyss failed to create server.  %s", error);
+            xmlrpc_faultf(envP, "Unable to create Abyss channel switch "
+                          "given the socket address.  %s", error);
             xmlrpc_strfree(error);
-        } else {
-            *chanSwitchPP = chanSwitchP;
-                    
-            ServerSetName(serverP, "XmlRpcServer");
-            
-            if (logFileName)
-                ServerSetLogFileName(serverP, logFileName);
         }
-        if (envP->fault_occurred)
-            ChanSwitchDestroy(chanSwitchP);
+    }
+}
+
+
+
+static void
+createChanSwitchIpv4Port(xmlrpc_env *          const envP,
+                         unsigned int          const portNumber,
+                         TChanSwitch **        const chanSwitchPP) {
+
+    struct sockaddr_in sockAddr;
+    const char * error;
+
+    sockAddr.sin_family      = AF_INET;
+    sockAddr.sin_port        = htons(portNumber);
+    sockAddr.sin_addr.s_addr = INADDR_ANY;
+
+    chanSwitchCreateSockAddr(PF_INET, (const struct sockaddr *)&sockAddr,
+                             sizeof(sockAddr),
+                             chanSwitchPP, &error);
+    
+    if (error) {
+        xmlrpc_faultf(envP, "Unable to create Abyss channel switch "
+                      "to listen on Port %u at any IPv4 address.  %s",
+                      portNumber, error);
+        xmlrpc_strfree(error);
     }
 }
 
@@ -1077,28 +646,48 @@ createServerBare(xmlrpc_env *                      const envP,
    to use.
 -----------------------------------------------------------------------------*/
     bool socketBound;
+    const struct sockaddr * sockAddrP;
+    socklen_t sockAddrLen;
     unsigned int portNumber;
     TOsSocket socketFd;
     const char * logFileName;
 
     extractServerCreateParms(envP, parmsP, parmSize,
-                             &socketBound, &portNumber, &socketFd,
+                             &socketBound,
+                             &sockAddrP, &sockAddrLen, &portNumber, &socketFd,
                              &logFileName);
 
     if (!envP->fault_occurred) {
+        TChanSwitch * chanSwitchP;
+
         if (socketBound)
-            createServerBoundSocket(envP, socketFd, logFileName,
-                                    serverP, chanSwitchPP);
+            createChanSwitchOsSocket(envP, socketFd, &chanSwitchP);
         else {
-            abyss_bool success;
+            if (sockAddrP)
+                createChanSwitchSockAddr(envP, sockAddrP, sockAddrLen,
+                                         &chanSwitchP);
+            else
+                createChanSwitchIpv4Port(envP, portNumber, &chanSwitchP);
+        }
+        if (!envP->fault_occurred) {
+            const char * error;
 
-            success = ServerCreate(serverP, "XmlRpcServer", portNumber,
-                                   DEFAULT_DOCS, logFileName);
+            ServerCreateSwitch(serverP, chanSwitchP, &error);
 
-            if (!success)
-                xmlrpc_faultf(envP, "Failed to create an Abyss server object");
+            if (error) {
+                xmlrpc_faultf(envP, "Abyss failed to create server.  %s",
+                              error);
+                xmlrpc_strfree(error);
+            } else {
+                *chanSwitchPP = chanSwitchP;
+                    
+                ServerSetName(serverP, "XmlRpcServer");
             
-            *chanSwitchPP = NULL;
+                if (logFileName)
+                    ServerSetLogFileName(serverP, logFileName);
+            }
+            if (envP->fault_occurred)
+                ChanSwitchDestroy(chanSwitchP);
         }
         if (logFileName)
             xmlrpc_strfree(logFileName);
@@ -1177,6 +766,8 @@ createServer(xmlrpc_env *                      const envP,
     createServerBare(envP, parmsP, parmSize, abyssServerP, chanSwitchPP);
 
     if (!envP->fault_occurred) {
+        const char * error;
+
         setAdditionalServerParms(parmsP, parmSize, abyssServerP);
         
         setHandlersRegistry(abyssServerP, uriPathParm(parmsP, parmSize),
@@ -1186,7 +777,12 @@ createServer(xmlrpc_env *                      const envP,
                             expiresParm(parmsP, parmSize),
                             maxAgeParm(parmsP, parmSize));
         
-        ServerInit(abyssServerP);
+        ServerInit2(abyssServerP, &error);
+
+        if (error) {
+            xmlrpc_faultf(envP, error);
+            xmlrpc_strfree(error);
+        }
     }
 }
 
@@ -1353,7 +949,7 @@ sigchld(int const signalClass ATTR_UNUSED) {
    forking as a threading mechanism), so we respond by passing the
    signal on to the Abyss server.  And reaping the dead child.
 -----------------------------------------------------------------------------*/
-#ifndef WIN32
+#if !MSVCRT
     /* Reap zombie children / report to Abyss until there aren't any more. */
 
     bool childrenLeft;
@@ -1380,7 +976,7 @@ sigchld(int const signalClass ATTR_UNUSED) {
         } else
             ServerHandleSigchld(pid);
     }
-#endif /* WIN32 */
+#endif /* MSVCRT */
 }
 
 
@@ -1390,11 +986,11 @@ struct xmlrpc_server_abyss_sig {
        functions in this library messed with them; useful for restoring
        them later.
     */
-#ifndef WIN32
+#if MSVCRT
+    int dummy;
+#else
     struct sigaction pipe;
     struct sigaction chld;
-#else
-    int dummy;
 #endif
 };
 
@@ -1402,7 +998,7 @@ struct xmlrpc_server_abyss_sig {
 
 static void
 setupSignalHandlers(xmlrpc_server_abyss_sig * const oldHandlersP) {
-#ifndef WIN32
+#if !MSVCRT
     struct sigaction mysigaction;
     
     sigemptyset(&mysigaction.sa_mask);
@@ -1422,8 +1018,7 @@ setupSignalHandlers(xmlrpc_server_abyss_sig * const oldHandlersP) {
 
 static void
 restoreSignalHandlers(const xmlrpc_server_abyss_sig * const oldHandlersP) {
-#ifndef WIN32
-
+#if !MSVCRT
     sigaction(SIGPIPE, &oldHandlersP->pipe, NULL);
     sigaction(SIGCHLD, &oldHandlersP->chld, NULL);
 
@@ -1583,7 +1178,7 @@ xmlrpc_server_abyss(xmlrpc_env *                      const envP,
                     const xmlrpc_server_abyss_parms * const parmsP,
                     unsigned int                      const parmSize) {
 /*----------------------------------------------------------------------------
-   Note that this is not re-entrant and not thread-safe, due to the
+   Note that this is not re-entrant and not thread-safe, because of the
    global library initialization.  If you want to run a server inside
    a thread of a multi-threaded program, use
    xmlrpc_server_abyss_create() instead.  As required by that
@@ -1608,7 +1203,6 @@ xmlrpc_server_abyss(xmlrpc_env *                      const envP,
                 oldHighLevelAbyssRun(envP, parmsP, parmSize);
             else
                 normalLevelAbyssRun(envP, parmsP, parmSize);
-            
         }
         xmlrpc_server_abyss_global_term();
     }
